@@ -18,6 +18,8 @@
  */
 
 #include <json-glib/json-glib.h>
+#include <click.h>
+#include <upstart.h>
 #include "helpers.h"
 
 /* Take an app ID and validate it and then break it up
@@ -61,62 +63,54 @@ gchar *
 manifest_to_desktop (const gchar * app_dir, const gchar * app_id)
 {
 	gchar * package = NULL;
-	gchar * output = NULL;
 	gchar * application = NULL;
 	gchar * version = NULL;
-	JsonParser * parser = NULL;
-	GError * error = NULL;
+	ClickDB * db = NULL;
+	ClickUser * user = NULL;
+	JsonObject * manifest = NULL;
 	gchar * desktoppath = NULL;
+	GError * error = NULL;
 
 	if (!app_id_to_triplet(app_id, &package, &application, &version)) {
 		g_warning("Unable to parse triplet: %s", app_id);
 		return NULL;
 	}
 
-	gchar * cmdline = g_strdup_printf("click info \"%s\"", package);
-	g_spawn_command_line_sync(cmdline, &output, NULL, NULL, &error);
-	g_free(cmdline);
-
+	db = click_db_new();
+	/* If TEST_CLICK_DB is unset, this reads the system database. */
+	click_db_read(db, g_getenv("TEST_CLICK_DB"), &error);
+	if (error != NULL) {
+		g_warning("Unable to read Click database: %s", error->message);
+		goto manifest_out;
+	}
+	/* If TEST_CLICK_USER is unset, this uses the current user name. */
+	user = click_user_new_for_user(db, g_getenv("TEST_CLICK_USER"), &error);
+	if (error != NULL) {
+		g_warning("Unable to read Click database: %s", error->message);
+		goto manifest_out;
+	}
+	manifest = click_user_get_manifest(user, package, &error);
 	if (error != NULL) {
 		g_warning("Unable to get manifest for '%s': %s", package, error->message);
-		g_error_free(error);
 		goto manifest_out;
 	}
 
-	parser = json_parser_new();
-
-	json_parser_load_from_data(parser, output, -1, &error);
-	g_free(output);
-
-	if (error != NULL) {
-		g_warning("Unable to load manifest data '%s': %s", package, error->message);
-		g_error_free(error);
-		goto manifest_out;
-	}
-
-	JsonNode * root = json_parser_get_root(parser);
-	if (json_node_get_node_type(root) != JSON_NODE_OBJECT) {
-		g_warning("Manifest '%s' doesn't start with an object", package);
-		goto manifest_out;
-	}
-
-	JsonObject * rootobj = json_node_get_object(root);
-	if (!json_object_has_member(rootobj, "version")) {
+	if (!json_object_has_member(manifest, "version")) {
 		g_warning("Manifest '%s' doesn't have a version", package);
 		goto manifest_out;
 	}
 
-	if (g_strcmp0(json_object_get_string_member(rootobj, "version"), version) != 0) {
-		g_warning("Manifest '%s' version '%s' doesn't match AppID version '%s'", package, json_object_get_string_member(rootobj, "version"), version);
+	if (g_strcmp0(json_object_get_string_member(manifest, "version"), version) != 0) {
+		g_warning("Manifest '%s' version '%s' doesn't match AppID version '%s'", package, json_object_get_string_member(manifest, "version"), version);
 		goto manifest_out;
 	}
 
-	if (!json_object_has_member(rootobj, "hooks")) {
+	if (!json_object_has_member(manifest, "hooks")) {
 		g_warning("Manifest '%s' doesn't have an hooks section", package);
 		goto manifest_out;
 	}
 
-	JsonObject * appsobj = json_object_get_object_member(rootobj, "hooks");
+	JsonObject * appsobj = json_object_get_object_member(manifest, "hooks");
 	if (appsobj == NULL) {
 		g_warning("Manifest '%s' has an hooks section that is not a JSON object", package);
 		goto manifest_out;
@@ -150,7 +144,12 @@ manifest_to_desktop (const gchar * app_dir, const gchar * app_id)
 	}
 
 manifest_out:
-	g_clear_object(&parser);
+	if (error != NULL)
+		g_error_free(error);
+	if (manifest != NULL)
+		json_object_unref(manifest);
+	g_clear_object(&user);
+	g_clear_object(&db);
 	g_free(package);
 	g_free(application);
 	g_free(version);
@@ -220,37 +219,71 @@ desktop_to_exec (GKeyFile * desktop_file, const gchar * from)
 
 /* Sets an upstart variable, currently using initctl */
 void
-set_upstart_variable (const gchar * variable, const gchar * value)
+set_upstart_variable (const gchar * variable, const gchar * value, gboolean sync)
 {
-	GError * error = NULL;
-	gchar * command[4] = {
-		"initctl",
-		"set-env",
-		NULL,
-		NULL
-	};
+	/* Check to see if we can get the job environment */
+	const gchar * job_name = g_getenv("UPSTART_JOB");
+	const gchar * instance_name = g_getenv("UPSTART_INSTANCE");
+	g_return_if_fail(job_name != NULL);
 
-	g_debug("Setting Upstart variable '%s' to '%s'", variable, value);
-	gchar * variablestr = g_strdup_printf("%s=%s", variable, value);
-	command[2] = variablestr;
+	/* Get a bus, let's go! */
+	GDBusConnection * bus = g_bus_get_sync(G_BUS_TYPE_SESSION, NULL, NULL);
+	g_return_if_fail(bus != NULL);
 
-	g_spawn_sync(NULL, /* working directory */
-		command,
-		NULL, /* environment */
-		G_SPAWN_SEARCH_PATH,
-		NULL, NULL, /* child setup */
-		NULL, /* stdout */
-		NULL, /* stderr */
-		NULL, /* exit status */
-		&error);
+	GVariantBuilder builder; /* Target: (assb) */
+	g_variant_builder_init(&builder, G_VARIANT_TYPE_TUPLE);
 
-	if (error != NULL) {
-		g_warning("Unable to set variable '%s' to '%s': %s", variable, value, error->message);
-		g_error_free(error);
+	/* Setup the job properties */
+	g_variant_builder_open(&builder, G_VARIANT_TYPE_ARRAY);
+	g_variant_builder_add_value(&builder, g_variant_new_string(job_name));
+	if (instance_name != NULL)
+		g_variant_builder_add_value(&builder, g_variant_new_string(instance_name));
+	g_variant_builder_close(&builder);
+
+	/* The value itself */
+	gchar * envstr = g_strdup_printf("%s=%s", variable, value);
+	g_variant_builder_add_value(&builder, g_variant_new_take_string(envstr));
+
+	/* Do we want to replace?  Yes, we do! */
+	g_variant_builder_add_value(&builder, g_variant_new_boolean(TRUE));
+
+	if (sync) {
+		GError * error = NULL;
+		GVariant * reply = g_dbus_connection_call_sync(bus,
+			DBUS_SERVICE_UPSTART,
+			DBUS_PATH_UPSTART,
+			DBUS_INTERFACE_UPSTART,
+			"SetEnv",
+			g_variant_builder_end(&builder),
+			NULL, /* reply */
+			G_DBUS_CALL_FLAGS_NONE,
+			-1, /* timeout */
+			NULL, /* cancelable */
+			&error); /* error */
+
+		if (reply != NULL) {
+			g_variant_unref(reply);
+		}
+
+		if (error != NULL) {
+			g_warning("Unable to set environment variable '%s' to '%s': %s", variable, value, error->message);
+			g_error_free(error);
+		}
+	} else {
+		g_dbus_connection_call(bus,
+			DBUS_SERVICE_UPSTART,
+			DBUS_PATH_UPSTART,
+			DBUS_INTERFACE_UPSTART,
+			"SetEnv",
+			g_variant_builder_end(&builder),
+			NULL, /* reply */
+			G_DBUS_CALL_FLAGS_NONE,
+			-1, /* timeout */
+			NULL, /* cancelable */
+			NULL, NULL); /* callback */
 	}
 
-	g_free(variablestr);
-	return;
+	g_object_unref(bus);
 }
 
 /* Convert a URI into a file */
@@ -273,133 +306,61 @@ uri2file (const gchar * uri)
 	return retval;
 }
 
-/* free a string in an array */
-static void
-free_string (gpointer value)
+/* Put the list of files into the argument array */
+static inline void
+file_list_handling (GArray * outarray, gchar ** list, gchar * (*dup_func) (const gchar * in))
 {
-	gchar ** str = (gchar **)value;
-	g_free(*str);
-	return;
-}
-
-/* Builds the file list from the URI list */
-static gchar *
-build_file_list (const gchar * uri_list)
-{
-	/* TODO: Joining only with space could cause issues with breaking them
-	   back out.  We don't have any cases of more than one today.  But, this
-	   isn't good.
-	   https://bugs.launchpad.net/upstart-app-launch/+bug/1229354
-	   */
-	gchar ** uri_split = g_strsplit(uri_list, " ", 0);
-
-	GArray * outarray = g_array_new(TRUE, FALSE, sizeof(gchar *));
-	g_array_set_clear_func(outarray, free_string);
+	/* No URLs, cool, this is a noop */
+	if (list == NULL || list[0] == NULL) {
+		return;
+	}
 
 	int i;
-	for (i = 0; uri_split[i] != NULL; i++) {
-		gchar * path = uri2file(uri_split[i]);
-		g_array_append_val(outarray, path);
+	for (i = 0; list[i] != NULL; i++) {
+		gchar * entry = dup_func(list[i]);
+
+		/* No NULLs */
+		if (entry != NULL && entry[0] != '\0') {
+			g_array_append_val(outarray, entry);
+		} else {
+			g_free(entry);
+		}
 	}
-
-	gchar * filelist = g_strjoinv(" ", (gchar **)outarray->data);
-	g_array_free(outarray, TRUE);
-
-	g_strfreev(uri_split);
-
-	return filelist;
-}
-
-/* Make sure we have the single URI variable */
-static inline void
-ensure_singleuri (gchar ** single_uri, const gchar * uri_list)
-{
-	if (uri_list == NULL) {
-		return;
-	}
-
-	if (*single_uri != NULL) {
-		return;
-	}
-
-	/* TODO: Joining only with space could cause issues with breaking them
-	   back out.  We don't have any cases of more than one today.  But, this
-	   isn't good.
-	   https://bugs.launchpad.net/upstart-app-launch/+bug/1229354
-	   */
-
-	gchar * first_uri = g_strdup(uri_list);
-	gchar * first_space = g_utf8_strchr(first_uri, -1, ' ');
-	
-	if (first_space != NULL) {
-		first_space[0] = '\0';
-	}
-
-	*single_uri = first_uri;
-
-	return;
-}
-
-/* Make sure we have a single file variable */
-static inline void
-ensure_singlefile (gchar ** single_file, const gchar * uri_list)
-{
-	if (uri_list == NULL) {
-		return;
-	}
-
-	if (*single_file != NULL) {
-		return;
-	}
-
-	/* TODO: Joining only with space could cause issues with breaking them
-	   back out.  We don't have any cases of more than one today.  But, this
-	   isn't good.
-	   https://bugs.launchpad.net/upstart-app-launch/+bug/1229354
-	   */
-
-	gchar * first_uri = g_strdup(uri_list);
-	gchar * first_space = g_utf8_strchr(first_uri, -1, ' ');
-	
-	if (first_space != NULL) {
-		first_space[0] = '\0';
-	}
-
-	gchar * first_file = NULL;
-	if (first_uri != NULL) {
-		first_file = uri2file(first_uri);
-	}
-
-	if (first_file != NULL) {
-		*single_file = first_file;
-	}
-
-	return;
 }
 
 /* Parse a desktop exec line and return the next string */
-static gchar *
-desktop_exec_segment_parse (const gchar * execline, const gchar * uri_list)
+static void
+desktop_exec_segment_parse (GArray * finalarray, const gchar * execsegment, gchar ** uri_list)
 {
-	gchar ** execsplit = g_strsplit(execline, "%", 0);
+	/* No NULL strings */
+	if (execsegment == NULL || execsegment[0] == '\0')
+		return;
+
+	/* Handle %F and %U as an argument on their own as per the spec */
+	if (g_strcmp0(execsegment, "%U") == 0) {
+		return file_list_handling(finalarray, uri_list, g_strdup);
+	}
+	if (g_strcmp0(execsegment, "%F") == 0) {
+		return file_list_handling(finalarray, uri_list, uri2file);
+	}
+
+	/* Start looking at individual codes */
+	gchar ** execsplit = g_strsplit(execsegment, "%", 0);
 
 	/* If we didn't have any codes, just exit here */
 	if (execsplit[1] == NULL) {
 		g_strfreev(execsplit);
-		return g_strdup(execline);
-	}
-
-	if (uri_list != NULL && uri_list[0] == '\0') {
-		uri_list = NULL;
+		gchar * dup = g_strdup(execsegment);
+		g_array_append_val(finalarray, dup);
+		return;
 	}
 
 	int i;
-	gchar * single_uri = NULL;
-	gchar * single_file = NULL;
-	gchar * file_list = NULL;
+
 	gboolean previous_percent = FALSE;
 	GArray * outarray = g_array_new(TRUE, FALSE, sizeof(const gchar *));
 	g_array_append_val(outarray, execsplit[0]);
+	gchar * single_file = NULL;
 
 	/* The variables allowed in an exec line from the Freedesktop.org Desktop
 	   File specification: http://standards.freedesktop.org/desktop-entry-spec/desktop-entry-spec-latest.html#exec-variables */
@@ -430,22 +391,16 @@ desktop_exec_segment_parse (const gchar * execline, const gchar * uri_list)
 			g_array_append_val(outarray, skipchar);
 			break;
 		case 'f':
-			ensure_singlefile(&single_file, uri_list);
-
-			if (single_file != NULL) {
+			if (uri_list != NULL && uri_list[0] != NULL) {
+				if (single_file == NULL)
+					single_file = uri2file(uri_list[0]);
 				g_array_append_val(outarray, single_file);
 			}
 
 			g_array_append_val(outarray, skipchar);
 			break;
 		case 'F':
-			if (uri_list != NULL) {
-				if (file_list == NULL) {
-					file_list = build_file_list(uri_list);
-				}
-				g_array_append_val(outarray, file_list);
-			}
-
+			g_warning("Exec line segment has a '%%F' that isn't its own argument '%s', ignoring.", execsegment);
 			g_array_append_val(outarray, skipchar);
 			break;
 		case 'i':
@@ -455,21 +410,12 @@ desktop_exec_segment_parse (const gchar * execline, const gchar * uri_list)
 			g_array_append_val(outarray, skipchar);
 			break;
 		case 'U':
-			/* TODO: Joining only with space could cause issues with breaking them
-			   back out.  We don't have any cases of more than one today.  But, this
-			   isn't good.
-			   https://bugs.launchpad.net/upstart-app-launch/+bug/1229354
-			   */
-			if (uri_list != NULL) {
-				g_array_append_val(outarray, uri_list);
-			}
+			g_warning("Exec line segment has a '%%U' that isn't its own argument '%s', ignoring.", execsegment);
 			g_array_append_val(outarray, skipchar);
 			break;
 		case 'u':
-			ensure_singleuri(&single_uri, uri_list);
-
-			if (single_uri != NULL) {
-				g_array_append_val(outarray, single_uri);
+			if (uri_list != NULL && uri_list[0] != NULL) {
+				g_array_append_val(outarray, uri_list[0]);
 			}
 
 			g_array_append_val(outarray, skipchar);
@@ -484,12 +430,14 @@ desktop_exec_segment_parse (const gchar * execline, const gchar * uri_list)
 	gchar * output = g_strjoinv(NULL, (gchar **)outarray->data);
 	g_array_free(outarray, TRUE);
 
-	g_free(single_uri);
-	g_free(single_file);
-	g_free(file_list);
-	g_strfreev(execsplit);
+	if (output != NULL && output[0] != '\0') {
+		g_array_append_val(finalarray, output);
+	} else {
+		g_free(output);
+	}
 
-	return output;
+	g_free(single_file);
+	g_strfreev(execsplit);
 }
 
 /* Take a full exec line, split it out, parse the segments and return
@@ -499,6 +447,7 @@ desktop_exec_parse (const gchar * execline, const gchar * urilist)
 {
 	GError * error = NULL;
 	gchar ** splitexec = NULL;
+	gchar ** splituris = NULL;
 	gint execitems = 0;
 
 	/* This returns from desktop file style quoting to straight strings with
@@ -513,102 +462,46 @@ desktop_exec_parse (const gchar * execline, const gchar * urilist)
 		return NULL;
 	}
 
+	if (urilist != NULL && urilist[0] != '\0') {
+		g_shell_parse_argv(urilist, NULL, &splituris, &error);
+
+		if (error != NULL) {
+			g_warning("Unable to parse URIs '%s': %s", urilist, error->message);
+			g_error_free(error);
+			/* Continuing without URIs */
+			splituris = NULL;
+		}
+	}
+
+
 	GArray * newargv = g_array_new(TRUE, FALSE, sizeof(gchar *));
 	int i;
 	for (i = 0; i < execitems; i++) {
-		gchar * execinserted = desktop_exec_segment_parse(splitexec[i], urilist);
-		g_array_append_val(newargv, execinserted);
+		desktop_exec_segment_parse(newargv, splitexec[i], splituris);
 	}
 	g_strfreev(splitexec);
 
+	if (splituris != NULL) {
+		g_strfreev(splituris);
+	}
+
 	/* Each string here should be its own param */
 
 	return newargv;
 }
 
-/* Check to make sure we have the sections and keys we want */
-static gboolean
-verify_keyfile (GKeyFile * inkeyfile, const gchar * desktop)
-{
-	if (inkeyfile == NULL) return FALSE;
-
-	if (!g_key_file_has_group(inkeyfile, "Desktop Entry")) {
-		g_warning("Desktop file '%s' is missing the 'Desktop Entry' group", desktop);
-		return FALSE;
-	}
-
-	if (!g_key_file_has_key(inkeyfile, "Desktop Entry", "Exec", NULL)) {
-		g_warning("Desktop file '%s' is missing the 'Exec' key", desktop);
-		return FALSE;
-	}
-
-	return TRUE;
-}
-
-/* Try to find a desktop file in a particular data directory */
-static GKeyFile *
-try_dir (const char * dir, const gchar * desktop)
-{
-	gchar * fullpath = g_build_filename(dir, "applications", desktop, NULL);
-	GKeyFile * keyfile = g_key_file_new();
-
-	/* NOTE: Leaving off the error here as we'll get a bunch of them,
-	   so individuals aren't really useful */
-	gboolean loaded = g_key_file_load_from_file(keyfile, fullpath, G_KEY_FILE_NONE, NULL);
-
-	g_free(fullpath);
-
-	if (!loaded) {
-		g_key_file_free(keyfile);
-		return NULL;
-	}
-
-	if (!verify_keyfile(keyfile, desktop)) {
-		g_key_file_free(keyfile);
-		return NULL;
-	}
-
-	return keyfile;
-}
-
-/* Find the keyfile that we need for a particular AppID and return it.
-   Or NULL if we can't find it. */
-GKeyFile *
-keyfile_for_appid (const gchar * appid, gchar ** desktopfile)
-{
-	gchar * desktop = g_strdup_printf("%s.desktop", appid);
-
-	const char * const * data_dirs = g_get_system_data_dirs();
-	GKeyFile * keyfile = NULL;
-	int i;
-
-	keyfile = try_dir(g_get_user_data_dir(), desktop);
-	if (keyfile != NULL && desktopfile != NULL && *desktopfile == NULL) {
-		*desktopfile = g_build_filename(g_get_user_data_dir(), "applications", desktop, NULL);
-	}
-
-	for (i = 0; data_dirs[i] != NULL && keyfile == NULL; i++) {
-		keyfile = try_dir(data_dirs[i], desktop);
-
-		if (keyfile != NULL && desktopfile != NULL && *desktopfile == NULL) {
-			*desktopfile = g_build_filename(data_dirs[i], "applications", desktop, NULL);
-		}
-	}
-
-	g_free(desktop);
-
-	return keyfile;
-}
-
 /* Set environment various variables to make apps work under
  * confinement according to:
  * https://wiki.ubuntu.com/SecurityTeam/Specifications/ApplicationConfinement
  */
 void
-set_confined_envvars (const gchar * package)
+set_confined_envvars (const gchar * package, const gchar * app_dir)
 {
+	g_return_if_fail(package != NULL);
+	g_return_if_fail(app_dir != NULL);
+
 	g_debug("Setting 'UBUNTU_APPLICATION_ISOLATION' to '1'");
-	set_upstart_variable("UBUNTU_APPLICATION_ISOLATION", "1");
+	set_upstart_variable("UBUNTU_APPLICATION_ISOLATION", "1", FALSE);
 
 	/* Make sure the XDG base dirs are set for the application using
 	 * the user's current values/system defaults. We could set these to
@@ -616,21 +509,26 @@ set_confined_envvars (const gchar * package)
 	 * brittle if someone uses different base dirs.
 	 */
 	g_debug("Setting 'XDG_CACHE_HOME' using g_get_user_cache_dir()");
-	set_upstart_variable("XDG_CACHE_HOME", g_get_user_cache_dir());
+	set_upstart_variable("XDG_CACHE_HOME", g_get_user_cache_dir(), FALSE);
 
 	g_debug("Setting 'XDG_CONFIG_HOME' using g_get_user_config_dir()");
-	set_upstart_variable("XDG_CONFIG_HOME", g_get_user_config_dir());
+	set_upstart_variable("XDG_CONFIG_HOME", g_get_user_config_dir(), FALSE);
 
 	g_debug("Setting 'XDG_DATA_HOME' using g_get_user_data_dir()");
-	set_upstart_variable("XDG_DATA_HOME", g_get_user_data_dir());
+	set_upstart_variable("XDG_DATA_HOME", g_get_user_data_dir(), FALSE);
 
 	g_debug("Setting 'XDG_RUNTIME_DIR' using g_get_user_runtime_dir()");
-	set_upstart_variable("XDG_RUNTIME_DIR", g_get_user_runtime_dir());
+	set_upstart_variable("XDG_RUNTIME_DIR", g_get_user_runtime_dir(), FALSE);
+
+	/* Add the application's dir to the list of sources for data */
+	gchar * datadirs = g_strjoin(":", app_dir, g_getenv("XDG_DATA_DIRS"), NULL);
+	set_upstart_variable("XDG_DATA_DIRS", datadirs, FALSE);
+	g_free(datadirs);
 
 	/* Set TMPDIR to something sane and application-specific */
 	gchar * tmpdir = g_strdup_printf("%s/confined/%s", g_get_user_runtime_dir(), package);
 	g_debug("Setting 'TMPDIR' to '%s'", tmpdir);
-	set_upstart_variable("TMPDIR", tmpdir);
+	set_upstart_variable("TMPDIR", tmpdir, FALSE);
 	g_debug("Creating '%s'", tmpdir);
 	g_mkdir_with_parents(tmpdir, 0700);
 	g_free(tmpdir);
@@ -638,8 +536,90 @@ set_confined_envvars (const gchar * package)
 	/* Do the same for nvidia */
 	gchar * nv_shader_cachedir = g_strdup_printf("%s/%s", g_get_user_cache_dir(), package);
 	g_debug("Setting '__GL_SHADER_DISK_CACHE_PATH' to '%s'", nv_shader_cachedir);
-	set_upstart_variable("__GL_SHADER_DISK_CACHE_PATH", nv_shader_cachedir);
+	set_upstart_variable("__GL_SHADER_DISK_CACHE_PATH", nv_shader_cachedir, FALSE);
 	g_free(nv_shader_cachedir);
 
 	return;
 }
+
+static void
+unity_signal_cb (GDBusConnection * con, const gchar * sender, const gchar * path, const gchar * interface, const gchar * signal, GVariant * params, gpointer user_data)
+{
+	GMainLoop * mainloop = (GMainLoop *)user_data;
+	g_main_loop_quit(mainloop);
+}
+
+struct _handshake_t {
+	GDBusConnection * con;
+	GMainLoop * mainloop;
+	guint signal_subscribe;
+	guint timeout;
+};
+
+static gboolean
+unity_too_slow_cb (gpointer user_data)
+{
+	handshake_t * handshake = (handshake_t *)user_data;
+	g_main_loop_quit(handshake->mainloop);
+	handshake->timeout = 0;
+	return G_SOURCE_REMOVE;
+}
+
+handshake_t *
+starting_handshake_start (const gchar *   app_id)
+{
+	GError * error = NULL;
+	handshake_t * handshake = g_new0(handshake_t, 1);
+
+	handshake->mainloop = g_main_loop_new(NULL, FALSE);
+	handshake->con = g_bus_get_sync(G_BUS_TYPE_SESSION, NULL, &error);
+
+	if (error != NULL) {
+		g_critical("Unable to connect to session bus: %s", error->message);
+		g_error_free(error);
+		g_free(handshake);
+		return NULL;
+	}
+
+	/* Set up listening for the unfrozen signal from Unity */
+	handshake->signal_subscribe = g_dbus_connection_signal_subscribe(handshake->con,
+		NULL, /* sender */
+		"com.canonical.UpstartAppLaunch", /* interface */
+		"UnityStartingSignal", /* signal */
+		"/", /* path */
+		app_id, /* arg0 */
+		G_DBUS_SIGNAL_FLAGS_NONE,
+		unity_signal_cb, handshake->mainloop,
+		NULL); /* user data destroy */
+
+	/* Send unfreeze to to Unity */
+	g_dbus_connection_emit_signal(handshake->con,
+		NULL, /* destination */
+		"/", /* path */
+		"com.canonical.UpstartAppLaunch", /* interface */
+		"UnityStartingBroadcast", /* signal */
+		g_variant_new("(s)", app_id),
+		&error);
+
+	/* Really, Unity? */
+	handshake->timeout = g_timeout_add_seconds(1, unity_too_slow_cb, handshake);
+
+	return handshake;
+}
+
+void
+starting_handshake_wait (handshake_t * handshake)
+{
+	if (handshake == NULL)
+		return;
+
+	g_main_loop_run(handshake->mainloop);
+
+	if (handshake->timeout != 0)
+		g_source_remove(handshake->timeout);
+	g_main_loop_unref(handshake->mainloop);
+	g_dbus_connection_signal_unsubscribe(handshake->con, handshake->signal_subscribe);
+	g_object_unref(handshake->con);
+
+	g_free(handshake);
+}
